package nl.uvt.locator;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.Nodes;
import nu.xom.Text;
import nu.xom.XPathContext;

import org.apache.log4j.*;

public class Locator extends HttpServlet {

	public static final String SOAPNamespace   = "http://schemas.xmlsoap.org/soap/envelope/";
	public static final String SRWNamespace    = "http://www.loc.gov/zing/srw/";
	public static final String diagNamespace   = "http://www.loc.gov/zing/srw/diagnostic/";

	static XPathContext sruContext = new XPathContext();

	private void fillContext(){
		sruContext.addNamespace("SOAP-ENV", SOAPNamespace);
		sruContext.addNamespace("srw", SRWNamespace);
		sruContext.addNamespace("diag", diagNamespace);
	}

	public static Properties sruDefaultProps = new Properties();
	public static Properties locatorConfig = new Properties();
	public static Properties httpConfig = new Properties();

	static public Logger logger;

	static private String [] prefixPaths;
	static String myName;
	static String myHostName;
	static String myVersion;
	static String myConfigName;
	static Set<String> knownCollections = null;
	static Set<String> knownDepots = null;

	private void setSRUDefaults(){
		sruDefaultProps.setProperty("operation", "searchRetrieve");
		sruDefaultProps.setProperty("version", "1.1");
		if ( locatorConfig.getProperty("sruMaximumRecords") == null )
			sruDefaultProps.setProperty("maximumRecords", "5");
		else
			sruDefaultProps.setProperty( "maximumRecords", 
					locatorConfig.getProperty("sruMaximumRecords"));
	}

	static public File getFile( String name ){
		return getFile( prefixPaths, name );
	}

	static	private File getFile( String[] prefixes, String name ){
		try {
			int i=0;
			while ( i < prefixes.length ){
				String fullName = prefixes[i] + File.separator + name;
				File file = new File( fullName );
				if ( file.exists() ){
					return file;
				}
				++i;
			}
		}
		catch ( Exception e ){
		}
		return null;
	}

	private void readConfig( String name, Properties p ){
		readConfig( name, p, false );
	}

	private void readMainConfig( String name, Properties p ){
		readConfig( name, p, true );
	}

	static Map<String,Long > modTimes = new HashMap<String,Long>();
	
	private void readConfig( String name, Properties result, Boolean main ){
		try {
			File configFile = getFile( prefixPaths, name );
			if ( configFile != null ){
				long moDate = configFile.lastModified();
				long oldDate = 0;
				if ( modTimes.containsKey(name) ){
					oldDate = modTimes.get(name);
				}
//				logger.debug("file " + name + " old date was: " + oldDate + " moDate=" + moDate );
				if ( moDate > oldDate ){
					logger.debug("(re)read configfile:" + name );
					FileInputStream is = new FileInputStream(configFile);
					result.clear();
					result.load( is );
					if ( main ){
						myConfigName = configFile.getAbsolutePath();
						knownCollections = null;
						knownDepots = null;
					}
					modTimes.put(name, moDate);
					logger.debug("config:" + name + "=" + result.toString());
				}
			}		
		}
		catch ( Exception e ){
			logger.error("readConfig problem: "+e.getMessage() );
		}
	}	

	public static boolean isKnownCollection( String collection ) {
		if ( knownCollections == null ){
			knownCollections = new HashSet<String>();
			String colls = locatorConfig.getProperty("knownCollections","");
			if ( colls.equals("") ){
				// hardwired default
				logger.warn("PLEASE FIX THIS: NO knownCollections found in configfile." );
				colls = "CBM, CIF, BRA, DISC, ECO, GES, JUR, LET, SOC, TFL";
				logger.warn("PLEASE FIX THIS: we use hardcoded default: " + colls );
			}
			String [] doiS = colls.split(",");
			for ( int i=0; i < doiS.length; ++i ){
				knownCollections.add(doiS[i].trim());
			}
			logger.debug("filled Known collections list: " + knownCollections.toString() );
		}
		return knownCollections.contains( collection );
	}

	public static boolean isDepot(String collection ) {
		if ( knownDepots == null ){
			knownDepots = new HashSet<String>();
			String colls = Locator.locatorConfig.getProperty("knownDepots","");
			if ( colls.equals("") ){
				// hardwired default
				Locator.logger.warn("PLEASE FIX THIS: NO knownDepots found in configfile." );
				colls = "CBM, KOD, TRE";
				Locator.logger.warn("PLEASE FIX THIS: we use hardcoded default: " + colls );
			}
			String [] doiS = colls.split(",");
			for ( int i=0; i < doiS.length; ++i ){
				knownDepots.add(doiS[i].trim());
			}
			Locator.logger.debug("filled Known depots list: " + knownDepots.toString() );
		}
		return knownDepots.contains( collection );
	}

	public void init() throws ServletException{
		myName = getServletContext().getInitParameter("artifactId" );
		myVersion = getServletContext().getInitParameter("version" );
		String prefixPath = getServletContext().getInitParameter("prefixPath" );
		if ( prefixPath.equals("") )
			throw new ServletException("no prefixPath in web.xml found" );
		prefixPaths = prefixPath.split(":");
		if ( prefixPaths.length == 0 ){
			throw new ServletException("no prefix paths for configuration found" );
		}
		String configName = null;
		File l4j = getFile( "log4j.properties");
		if ( l4j != null){
			configName = l4j.getAbsolutePath();
			PropertyConfigurator.configure( configName );
		}
		else {
			BasicConfigurator.configure();
		}
		logger = Logger.getLogger(Locator.class);
		logger.info("initialize Locator, using config from PATH=" + prefixPath );
		if ( configName != null )
			logger.debug("log4j configured from " + configName );
		else
			logger.warn("couldn't get log4j.properties. logging to System.out");
		fillContext();
		logger.info("started " + myName + "-" + myVersion );
	}

	static boolean doMeresco = false;
	static boolean cascadingSearch = false;
	
	enum serveAction {unknownAct, infoAct, explainAct, lookupAct };

	public void doGet( HttpServletRequest request, 
			HttpServletResponse response ) throws ServletException, IOException {
		myHostName = request.getServerName();
		OutputStream os = response.getOutputStream();
		PrintStream out = new PrintStream( os, true, "UTF8" );
		String fileName = "locator.cfg";
		readMainConfig( fileName, locatorConfig );
		if ( locatorConfig.isEmpty() ){
			out.println("unable to read configfile " + fileName );
			logger.error("reading configfile '" + fileName + "' failed");
			return;
		}
		String httpSettings = locatorConfig.getProperty("httpDefaults");
		if ( httpSettings != null ){
			fileName = httpSettings;
			readConfig(fileName, httpConfig );
			if ( httpConfig.isEmpty() ){
				logger.warn("reading http configfile '" + fileName  + "' failed. (using defaults)");
			}
		}
		setSRUDefaults();
		response.setContentType("text/XML");
		String prop = locatorConfig.getProperty("doMeresco", "false");
		if ( prop.compareToIgnoreCase("true") == 0 ){
			doMeresco = true;
		}
		else
			doMeresco = false;
		prop = locatorConfig.getProperty("cascadingSearch", "false");
		if ( prop.compareToIgnoreCase("true") == 0 ){
			cascadingSearch = true;
		}
		else
			cascadingSearch = false;
		Document doc = null;
		String action = request.getParameter( "action" ); 	 
		serveAction act = serveAction.lookupAct; 	
		if ( action != null ){ 	 
			logger.debug("special_action="+action); 	 
			if ( action.equals("info") ) 	 
				act = serveAction.infoAct; 	 
			else if ( action.equals("explain") ) 	 
				act = serveAction.explainAct; 	 
			else if ( action.equals("meresco") ){
				doMeresco = true;
			}
			else if ( action.equals("cascading") ){
				cascadingSearch = true;
			}
			else  {
				doc = errorDoc( "invalid action:" + action );
			}
		}
		if ( doc == null ){
			if ( act == serveAction.explainAct ){
				doc = explainDoc();
			}
			else if ( act == serveAction.infoAct ){
				ServeData sD = new ServeData();
				if ( sD.parseQuery( request ) ){
					logger.debug( "parsed query: " + sD );
					doc = tryToServeDoc( sD );
				}
				else
					doc = errorDoc("invalid parameters");
			}
			else {
				QueryData pQ = new QueryData();
				try {
					if ( pQ.parseQuery( request ) ){
						logger.debug( "parsed query: " + pQ );
						if ( pQ.genre.equals("article") || pQ.genre.equals("journal"))
							doc = tryToLocateArticle( pQ );
						else
							doc = tryToLocateBook( pQ );
					}
					else
						doc = errorDoc("parseQuery failed");
				}
				catch (LocatorException e) {
					doc = errorDoc( 0, "query parsing failed", e.getMessage());
				}
			}
		}
//		logger.debug("resulting Doc: " + doc.toXML());
		out.println( doc.toXML() );		
	}

	static DocServer dS = null;

	private Document tryToServeDoc( ServeData sD ) {
		if ( dS == null)
			dS = new DocServer();
		Document doc = dS.refresh( sD );
		if ( doc == null )
			doc = dS.serverDoc( sD );
		return doc;
	}

	private Document tryToLocateArticle( QueryData pQ ){
		List<LinkInfo> linkInfo = null;
		try {
			LinkDbConnection linkDb = new LinkDbConnection();
			linkInfo = linkDb.query( pQ );
		}
		catch ( LocatorException e ){
			logger.warn(e.getMessage());
			return errorDoc( 0, "linkDb failed", e.getMessage() );
		}
		if ( linkInfo != null ){
			for ( Iterator<LinkInfo> it=linkInfo.iterator(); it.hasNext(); ){
				LinkInfo li = it.next();
				try{
					if ( !li.inRange( pQ )){
						it.remove();
					}
					else if ( !li.applyMethod( pQ ) ){
						it.remove();
					}
				}
				catch (LocatorException e ){
					return errorDoc( 0, "threshold failed", e.getMessage() );
				}
			}
		}
		List<LinkInfo> tuInfo = null;
		if ( cascadingSearch || linkInfo == null || linkInfo.isEmpty() ){
			if ( doMeresco ){
				if ( cascadingSearch )
					logger.debug("cascading search continues with meresco");
				else
					logger.debug("no valid linkdb results, try meresco");
				try {
					MerescoConnection bzv = new MerescoConnection();
					tuInfo = bzv.query( pQ );
				} catch (LocatorException e) {
					return errorDoc( 0, "meresco failed", e.getMessage() );
				}				
			}
			else {
				if ( cascadingSearch )
					logger.debug("cascading search continues with tucat");
				else
					logger.debug("no valid linkdb results, try tucat");
				try {
					TucatConnection tucat = new TucatConnection();
					tuInfo = tucat.query( pQ );
				} catch (LocatorException e) {
					return errorDoc( 0, "tucat failed", e.getMessage() );
				}
			}
		}
		if ( tuInfo != null ){
			for ( Iterator<LinkInfo> it=tuInfo.iterator(); it.hasNext(); ){
				LinkInfo li = it.next();
				if ( !li.applyMethod( pQ ) ){
					it.remove();
				}
			}
			if (linkInfo != null )
				linkInfo.addAll( tuInfo );
			else
				linkInfo = tuInfo;
		}
		boolean suc6 = false;
		if ( linkInfo != null ){
			for ( Iterator<LinkInfo> it=linkInfo.iterator(); it.hasNext() && !suc6; ){
				LinkInfo li = it.next();
				suc6 = li.diagnostics.isEmpty();
			}
		}
		if ( !suc6 )
			return IBLdoc( pQ, linkInfo );
		else
			return toDoc(linkInfo );
	}

	private Document tryToLocateBook( QueryData pQ ){
		if ( pQ.database == null ){
			return errorDoc( "missing 'database=' in query");
		}
		else {
			List<LinkInfo> linkInfo = null;
			if ( doMeresco ){
				try {
					MerescoConnection bzv = new MerescoConnection();
					linkInfo = bzv.query( pQ );
				} catch (LocatorException e) {
					return errorDoc( 0, "meresco failed", e.getMessage() );
				}				
			}
			else {
				try {
					TucatConnection tucat = new TucatConnection();
					linkInfo = tucat.query( pQ );
				} catch (LocatorException e) {
					return errorDoc( 0, "tucat failed", e.getMessage() );
				}
			}
			if ( linkInfo != null ){
				for ( Iterator<LinkInfo> it=linkInfo.iterator(); it.hasNext(); ){
					LinkInfo li = it.next();
					if ( !li.applyMethod( pQ ) ){
						it.remove();
					}
				}
			}
			if ( linkInfo != null && !linkInfo.isEmpty() ){
				return toDoc(linkInfo );
			}
			else {
				return IBLdoc( pQ, null );		
			}
		}
	}

	private Document IBLdoc( QueryData qD, List<LinkInfo> linkInfoList ) {
		IllRequestor illRequestor = new IllRequestor();
		LinkInfo li;
		if ( linkInfoList == null || linkInfoList.size() ==0  )
			li = new LinkInfo(); //empty stub
		else
			li = linkInfoList.get(0);
		Document doc = illRequestor.getDocument( qD, li ); 
		return doc;
	}

	public Document toDoc( List<LinkInfo> linkInfo ){
		Element root = new Element("records");
		Document doc = new Document( root );
		if ( linkInfo != null ){
			for ( int i=0; i< linkInfo.size(); ++i ){
				Node liNode = linkInfo.get(i).toRecordNode();
				root.appendChild( liNode );
			}
		}
		return doc;
	}

	static public int extractHits( Document doc ){
		Nodes numOfRecords = doc.query("//srw:numberOfRecords", sruContext);
		if( numOfRecords.size() > 0 ){
			Element numHitsElement = (Element) numOfRecords.get(0);
			String hitString = numHitsElement.getValue();
			int hits=Integer.parseInt(hitString);
			return hits;
		}
		return 0;
	}



	public static Document explainDoc( ) {
		Element explainNode = new Element("explain");
		Element subNode = new Element("artifactId");
		Text txt = new Text( myName );
		subNode.appendChild( txt );
		explainNode.appendChild( subNode);
		subNode = new Element("version");
		txt = new Text( myVersion );
		subNode.appendChild( txt );
		explainNode.appendChild( subNode);
		subNode = new Element("hostname");
		txt = new Text( myHostName );
		subNode.appendChild( txt );
		explainNode.appendChild( subNode);
		subNode = new Element("configFile");
		txt = new Text( myConfigName );
		subNode.appendChild( txt );
		explainNode.appendChild( subNode);
		return new Document( explainNode );
	}

	public static Document errorDoc( int code, String mess, String addI ) {
		Diagnostic diag = new Diagnostic( code, mess, addI );
		Element diagNode = diag.toElement();
		return new Document( diagNode );
	}

	public static Document errorDoc( String mess ) {
		return errorDoc( 1, mess, "" );
	}

	public Document notFoundPage( String what ) {
		return errorDoc( 2, "no hits", what );
	}	  


	static final long serialVersionUID=1;
}
